#! /usr/bin/env python3

import sys

try:
    import gi
except ImportError:
    sys.exit(77) # skip this test, gi module is not available

gi.require_version('Gtk', '4.0')

from gi.repository import Gtk

assert isinstance(Gtk.INVALID_LIST_POSITION, int), 'Gtk.INVALID_LIST_POSITION is not an int'



#!/usr/bin/env python3

import os
import subprocess
import sys

repodir = sys.argv[1]
profile = sys.argv[2]

sys.stdout.write("/* This file is auto-generated. Do not edit. */\n")
sys.stdout.write("#pragma once\n")
sys.stdout.write("\n")
sys.stdout.write(f"#define PROFILE \"{profile}\"\n")

short_sha = os.environ.get('CI_COMMIT_SHORT_SHA')
if short_sha is None:
    cmd = ["git", "-C", repodir, "rev-parse", "--short", "HEAD"]
    try:
        with subprocess.Popen(cmd, stdout=subprocess.PIPE) as p:
            short_sha = p.stdout.read().decode('utf-8').rstrip("\n")
    except FileNotFoundError:
        short_sha = ''
        if profile != 'default':
            short_sha = 'devel'

sys.stdout.write(f"#define VCS_TAG \"{short_sha}\"\n")



#!/usr/bin/env python3

import os
import shutil
import subprocess

from pathlib import PurePath


stylesheets = [
    'gtk/theme/Default/Default-light.css',
    'gtk/theme/Default/Default-dark.css',
    'gtk/theme/Default/Default-hc.css',
    'gtk/theme/Default/Default-hc-dark.css',
]

references = [
    'docs/reference/gtk/gtk4',
    'docs/reference/gsk/gsk4',
    'docs/reference/gdk/gdk4',
    'docs/reference/gdk/gdk4-wayland',
    'docs/reference/gdk/gdk4-x11',
]

sourceroot = os.environ.get('MESON_SOURCE_ROOT')
buildroot = os.environ.get('MESON_BUILD_ROOT')
distroot = os.environ.get('MESON_DIST_ROOT')

for stylesheet in stylesheets:
    stylesheet_path = PurePath(stylesheet)
    src = PurePath(sourceroot, stylesheet_path.with_suffix('.scss'))
    dst = PurePath(distroot, stylesheet_path)
    subprocess.call(['sassc', '-a', '-M', '-t', 'compact', src, dst])

for reference in references:
    src_path = os.path.join(buildroot, reference)
    if os.path.isdir(src_path):
        dst_path = os.path.join(distroot, reference)
        shutil.copytree(src_path, dst_path)



#!/usr/bin/env python3

from dataclasses import dataclass
from enum import Enum
import argparse
import re
import sys
import os

class Premultiplied(Enum):
    PREMULTIPLIED = 0,
    STRAIGHT = 1,
    ARGUMENT = 2

    def from_string(val):
        if val.lower() in [ 'true', 'premultiplied' ]:
            return Premultiplied.PREMULTIPLIED;
        elif val.lower() in [ 'false', 'straight' ]:
            return Premultiplied.STRAIGHT;
        elif val.lower() in [ 'argument' ]:
            return Premultiplied.ARGUMENT;
        else:
            raise Exception ('Expected "premultiplied" or "straight" or "argument"')

    def to_c_code(self, argument_name):
        if self == Premultiplied.PREMULTIPLIED:
            return 'TRUE'
        elif self == Premultiplied.STRAIGHT:
            return 'FALSE'
        elif self == Premultiplied.ARGUMENT:
            return argument_name
        else:
            raise Exception('FIXME: Figure out assert_not_reached() with Python')

@dataclass
class VarType:
    c_name: str
    glsl_name: str
    glsl_prefix: str
    vk_type: str
    gl_attrib_pointer_str: str

    def get_gl_type (self, size):
        if size == 1:
            return self.glsl_name
        elif size == 2:
            return self.glsl_prefix + 'vec2'
        elif size == 3:
            return self.glsl_prefix + 'vec3'
        elif size == 4:
            return self.glsl_prefix + 'vec4'
        elif size == 8:
            return self.glsl_prefix + 'mat2x4'
        elif size == 12:
            return self.glsl_prefix + 'mat3x4'
        elif size == 16:
            return self.glsl_prefix + 'mat4'
        else:
            raise Exception(f'''Invalid size {size} for {self}''')

    def gl_attrib_pointer (self, indent, location, size, n_attrs, offset):
        return self.gl_attrib_pointer_str.format (indent, location, size, n_attrs, offset)

VarType.FLOAT = VarType (
        c_name = 'float',
        glsl_name = 'float',
        glsl_prefix ='',
        vk_type = 'SFLOAT',
        gl_attrib_pointer_str = '{0}glVertexAttribPointer ({1},\n'
                                '{0}                       {2},\n'
                                '{0}                       GL_FLOAT,\n'
                                '{0}                       GL_FALSE,\n'
                                '{0}                       {3},\n'
                                '{0}                       GSIZE_TO_POINTER (offset + {4}));'
    )
VarType.INT = VarType (
        c_name = 'gint32',
        glsl_name ='int',
        glsl_prefix = 'i',
        vk_type = 'SINT',
        gl_attrib_pointer_str = '{0}glVertexAttribIPointer ({1},\n'
                                '{0}                        {2},\n'
                                '{0}                        GL_INT,\n'
                                '{0}                        {3},\n'
                                '{0}                        GSIZE_TO_POINTER (offset + {4}));'
    )
VarType.UINT = VarType (
        c_name = 'guint32',
        glsl_name = 'uint',
        glsl_prefix = 'u',
        vk_type = 'UINT',
        gl_attrib_pointer_str = '{0}glVertexAttribIPointer ({1},\n'
                                '{0}                        {2},\n'
                                '{0}                        GL_UNSIGNED_INT,\n'
                                '{0}                        {3},\n'
                                '{0}                        GSIZE_TO_POINTER (offset + {4}));'
    )

@dataclass
class Type:
    type: str
    pointer: bool
    var_type: VarType
    size: int
    struct_init: str

    def struct_initializer (self, indent, instance_name, var_name, struct_member, offset):
        return self.struct_init.format (indent, instance_name, var_name, struct_member, offset);

@dataclass
class VariationType:
    type: str
    bits: int
    glsl_string: str
    c_flag: str

    def to_glsl (self, offset_bits):
        return self.glsl_string.format (offset_bits)

    def to_c_flag (self, name, offset_bits):
        return self.c_flag.format (name, offset_bits)

@dataclass
class Input:
    name: str
    type: Type

@dataclass
class Attribute:
    name: str
    inputs: Input
    var_type: VarType
    size: int
    location: int

@dataclass
class Variable:
    name: str
    type: Type

@dataclass
class Variation:
    name: str
    offset_bits: int
    type: VariationType

@dataclass
class File:
    filename: str
    name: str
    var_name: str
    struct_name: str
    n_textures: int
    n_instances: int
    ccs_premultiplied: Premultiplied
    acs_premultiplied: Premultiplied
    acs_equals_ccs: bool
    opacity: bool
    dual_blend: bool
    variables: list[Variable]
    variations: list[Variation]

types = [
Type(
    type = 'float',
    pointer = False,
    var_type = VarType.FLOAT,
    size = 1,
    struct_init = '{0}{1}{3}[{4}] = {2};'
),
Type(
    type = 'guint32',
    pointer = False,
    var_type = VarType.UINT,
    size = 1,
    struct_init = '{0}{1}{3}[{4}] = {2};'
),
Type(
    type = 'graphene_point_t',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 2,
    struct_init = '{0}{1}{3}[{4}] = {2}->x + pass->offset.x;\n'
                  '{0}{1}{3}[{4} + 1] = {2}->y + pass->offset.y;'
),
Type(
    type = 'graphene_size_t',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 2,
    struct_init = '{0}{1}{3}[{4}] = {2}->width;\n'
                  '{0}{1}{3}[{4} + 1] = {2}->height;'
),
Type(
    type = 'graphene_vec2_t',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 2,
    struct_init = '{0}graphene_vec2_to_float ({2}, {1}{3});'
),
Type(
    type = 'graphene_vec4_t',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 4,
    struct_init = '{0}graphene_vec4_to_float ({2}, {1}{3});'
),
Type(
    type = 'graphene_rect_t',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 4,
    struct_init = '{0}gsk_gpu_rect_to_float ({2}, &pass->offset, {1}{3});'
),
Type(
    type = 'GskRoundedRect',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 12,
    struct_init = '{0}gsk_rounded_rect_to_float ({2}, &pass->offset, {1}{3});'
),
Type(
    type = 'graphene_matrix_t',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 16,
    struct_init = '{0}graphene_matrix_to_float ({2}, {1}{3});'
),
Type(
    type = 'GdkColor',
    pointer = True,
    var_type = VarType.FLOAT,
    size = 4,
    struct_init = '{0}gdk_color_to_float ({2}, acs, {1}{3});\n'
                  '{0}{1}{3}[3] *= pass->opacity;'
),
]

variation_types = [
VariationType(
    type = 'gboolean',
    bits = 1,
    glsl_string = '(((GSK_VARIATION >> {0}u) & 1u) == 1u)',
    c_flag = '(({0} & 1) << {1})'
),
VariationType(
    type = 'GdkBuiltinColorStateId',
    bits = 8,
    glsl_string = '((GSK_VARIATION >> {0}u) & 255u)',
    c_flag = '(({0} & 0xFF) << {1})'
),
VariationType(
    type = 'GskBlendMode',
    bits = 8,
    glsl_string = '((GSK_VARIATION >> {0}u) & 255u)',
    c_flag = '(({0} & 0xFF) << {1})'
),
VariationType(
    type = 'GskMaskMode',
    bits = 8,
    glsl_string = '((GSK_VARIATION >> {0}u) & 255u)',
    c_flag = '(({0} & 0xFF) << {1})'
),
VariationType(
    type = 'GskPorterDuff',
    bits = 8,
    glsl_string = '((GSK_VARIATION >> {0}u) & 255u)',
    c_flag = '(({0} & 0xFF) << {1})'
),
VariationType(
    type = 'GskRepeat',
    bits = 4,
    glsl_string = '((GSK_VARIATION >> {0}u) & 15u)',
    c_flag = '(({0} & 0xF) << {1})'
),
]

def strtobool (val):
    if val.lower() in [ 'true' ]:
        return True;
    elif val.lower() in [ 'false' ]:
        return False;
    else:
        raise Exception ('Expected "true" or "false"')

def read_file (filename):
    lines = open(filename).readlines()
    variables = []
    variations = []
    variation_bits = 0
    on = False
    n_textures = 0
    n_instances = 6
    ccs_premultiplied = Premultiplied.PREMULTIPLIED
    acs_premultiplied = Premultiplied.STRAIGHT
    acs_equals_ccs = False
    opacity = True
    dual_blend = False
    name = ''
    var_name = ''
    struct_name = ''

    for pos, line in enumerate (lines):
        line = line.strip ()
        try:
            if not on and line == "#ifdef GSK_PREAMBLE":
                on = True
            elif not on:
                pass
            elif line == "#endif /* GSK_PREAMBLE */":
                on = False
            elif line == "":
                pass
            elif match := re.search (r'^variation\s*:\s*(\w+)\s+(\w+)\s*;$', line):
                var_type = next ((x for x in variation_types if x.type == match.group (1)), None)
                if not var_type:
                    raise Exception (f'''Unknown variation type "{match.group (1)}"''')
                if variation_bits + var_type.bits > 32:
                    raise Exception (f'''too many bits taken by variations''')
                variations.append (Variation (name = match.group (2),
                                              offset_bits = variation_bits,
                                              type = var_type))
                variation_bits += var_type.bits
            elif match := re.search (r'(\w+)\s+(\w+)\s*;', line):
                var_type = next ((x for x in types if x.type == match.group (1)), None)
                if not var_type:
                    raise Exception (f'''Unknown type name "{match.group (1)}"''')
                variables.append (Input (name = match.group (2),
                                         type = var_type))
            elif match := re.search (r'^textures\s*=\s*(\d+)\s*;$', line):
                n_textures = int (match.group(1))
                if n_textures < 0 or n_textures > 2:
                    raise Exception (f'''Number of textures must be <= 2''')
            elif match := re.search (r'^instances\s*=\s*(\d+)\s*;$', line):
                n_instances = int (match.group(1))
            elif match := re.search (r'^name\s*=\s*"(\w+)"\s*;$', line):
                name = match.group(1)
            elif match := re.search (r'^var_name\s*=\s*"(\w+)"\s*;$', line):
                var_name = match.group(1)
            elif match := re.search (r'^struct_name\s*=\s*"(\w+)"\s*;$', line):
                struct_name = match.group(1)
            elif match := re.search (r'^ccs_premultiplied\s*=\s*(\w+)\s*;$', line):
                ccs_premultiplied = Premultiplied.from_string (match.group(1))
            elif match := re.search (r'^acs_premultiplied\s*=\s*(\w+)\s*;$', line):
                acs_premultiplied = Premultiplied.from_string (match.group(1))
            elif match := re.search (r'^acs_equals_ccs\s*=\s*(\w+)\s*;$', line):
                acs_equals_ccs = strtobool (match.group(1))
            elif match := re.search (r'^opacity\s*=\s*(\w+)\s*;$', line):
                opacity = strtobool (match.group(1))
            elif match := re.search (r'^dual_blend\s*=\s*(\w+)\s*;$', line):
                dual_blend = strtobool (match.group(1))
            else:
                raise Exception ('Could not parse line')
        except Exception as e:
            raise Exception (f'''{filename}:{pos}: {repr (e)}''')
    
    if not name:
        name = os.path.splitext(os.path.splitext(os.path.basename(filename))[0])[0][6:]
    if not var_name:
        var_name = "gsk_gpu_" + name.replace('-', '_')
    if not struct_name:
        struct_name = "GskGpu" + name.title().replace('-', '')

    return File (filename = filename,
                 name = name,
                 var_name = var_name,
                 struct_name = struct_name,
                 n_textures = n_textures,
                 n_instances = n_instances,
                 ccs_premultiplied = ccs_premultiplied,
                 acs_premultiplied = acs_premultiplied,
                 acs_equals_ccs = acs_equals_ccs,
                 opacity = opacity,
                 dual_blend = dual_blend,
                 variables = variables,
                 variations = variations)


def generate_attributes(inputs):
    inputs = inputs.copy ()
    result = []
    location = 0
    while inputs:
        input = inputs.pop (0)
        size = input.type.size
        var_type = input.type.var_type
        used = [input]
        while size < 4 and inputs:
            input = next ((i for i in inputs if i.type.size + size <= 4 and i.type.var_type == var_type), None)
            if not input:
                break
            inputs.remove (input)
            size += input.type.size
            used.append (input)

        attr = Attribute (name = '_'.join (x.name for x in used),
                          var_type = var_type,
                          size = size,
                          inputs = used,
                          location = location)
        location += (size + 3) // 4;
        result.append (attr)

    if location > 16:
        raise Exception (f'''MAX_VERTEX_ATTRIBS is 16, this shader uses {location}''')
            
    return location, result


def print_glsl_variations (variations):
    for var in variations:
        print (f'''#define VARIATION_{var.name.upper()} {var.type.to_glsl (var.offset_bits)}''')
    if variations:
        print ()

def print_glsl_attributes (attributes):
    print (f'''#ifdef GSK_VERTEX_SHADER''')
    for attr in attributes:
        print (f'''IN({attr.location}) {attr.var_type.get_gl_type (attr.size)} in_{attr.name};''')
        if len (attr.inputs) > 1:
            size = 0
            for input in attr.inputs:
                print (f'''#define in_{input.name} in_{attr.name}.{'xyzw'[size:size + input.type.size]}''')
                size += input.type.size
    print (f'''#endif''')

def print_c_struct (file, n_attributes, attributes):
    print (f'''typedef struct _{file.struct_name}Instance {file.struct_name}Instance;

struct _{file.struct_name}Instance {{''');

    for attr in attributes:
        print (f'''  {attr.var_type.c_name} {attr.name}[{max (4, attr.size)}];''')
        
    print (f'''}};
''')

@dataclass
class FunctionArg:
    type: str
    pointer: bool
    name: str

def print_c_function (retval, name, args, is_prototype):
    indent = len (name) + 2
    type_len = max (map (lambda arg: len (arg.type), args))
    
    print (retval)
    if args:
        print (f'''{name} ({args[0].type.ljust(type_len)} {'*' if args[0].pointer else ' '}{args[0].name}{',' if len (args) > 1 else (');' if is_prototype else ')')}''')
    else:
        print (f'''{name} (void){';' if is_prototype else ''}''')

    for pos, arg in enumerate (args[1:], 2):
        print (f'''{' '.ljust(indent)}{arg.type.ljust(type_len)} {'*' if arg.pointer else ' '}{arg.name}{',' if pos < len (args) else (');' if is_prototype else ')')}''')

def print_c_print_function (file):
    print_c_function ('static void',
                      file.var_name + '_op_print_instance',
                      [ FunctionArg ('GskGpuShaderOp', True,  'shader'),
                        FunctionArg ('gpointer',       False, 'instance_'),
                        FunctionArg ('GString',        True,  'string') ],
                      False)
    print (f'''{{
''')
    for var in file.variables:
        if var.name == 'bounds':
            print (f'''{file.struct_name}Instance *instance = instance_;

  gsk_gpu_print_rect (string, instance->bounds);
''')

    print (f'''  /* FIXME: Implement */
}}
''')

def print_c_shader_op_class (file):
    print (f'''static const GskGpuShaderOpClass {file.var_name.upper()}_OP_CLASS = {{
  {{
    GSK_GPU_OP_SIZE (GskGpuShaderOp),
    GSK_GPU_STAGE_SHADER,
    gsk_gpu_shader_op_finish,
    gsk_gpu_shader_op_print,
#ifdef GDK_RENDERING_VULKAN
    gsk_gpu_shader_op_vk_command,
#endif
    gsk_gpu_shader_op_gl_command
  }},
  "gskgpu{file.name}",
  {file.n_textures},
  {file.n_instances},
  sizeof ({file.struct_name}Instance),
#ifdef GDK_RENDERING_VULKAN
  &{file.var_name}_info,
#endif
  {file.var_name}_op_print_instance,
  {file.var_name}_setup_attrib_locations,
  {file.var_name}_setup_vao
}};
''')

def print_c_invocation (file, n_attributes, attributes, prototype_only):
    args = [ FunctionArg ('GskGpuRenderPass',            True,  'pass'),
             FunctionArg ('GdkColorState',               True,  'ccs') ]
    if file.ccs_premultiplied == Premultiplied.ARGUMENT:
        args.append (FunctionArg ('gboolean',            False,  'ccs_premultiplied'))
    if not file.acs_equals_ccs:
        args.append (FunctionArg ('GdkColorState',       True,  'acs'))
    if file.acs_premultiplied == Premultiplied.ARGUMENT:
        args.append (FunctionArg ('gboolean',            False,  'acs_premultiplied'))
    args.append (FunctionArg ('const graphene_rect_t',   True,  'bounds'))
    for i in range(1, file.n_textures + 1):
        args += [ FunctionArg ('GskGpuImage',             True, 'image' + str (i)),
                  FunctionArg ('GskGpuSampler',           False, 'sampler' + str (i)) ]

    for var in file.variations:
        args.append (FunctionArg (var.type.type, False, 'variation_' + var.name))
    for var in file.variables:
        if var.name == 'opacity' or var.name == 'bounds':
            continue
        args.append (FunctionArg (('const ' if var.type.pointer else '') + var.type.type, var.type.pointer, var.name))

    print_c_function ('void',
                      f'''{file.var_name}_op''',
                      args,
                      prototype_only)

    if prototype_only:
        return

    print (f'''{{
  {file.struct_name}Instance *instance;

  gsk_gpu_render_pass_prepare_shader (pass);

  gsk_gpu_shader_op_alloc (pass->frame,
                           &{file.var_name.upper()}_OP_CLASS,
                           ccs ? gsk_gpu_color_states_create (ccs, {file.ccs_premultiplied.to_c_code('ccs_premultiplied')}, {'ccs' if file.acs_equals_ccs else 'acs'}, {file.acs_premultiplied.to_c_code('acs_premultiplied')})
                               : gsk_gpu_color_states_create_equal ({file.ccs_premultiplied.to_c_code('ccs_premultiplied')}, {file.acs_premultiplied.to_c_code('acs_premultiplied')}),''')
    if file.variations:
        for pos, var in enumerate (file.variations, 1):
            print (f'''                           {var.type.to_c_flag ('variation_' + var.name, var.offset_bits)}{' |' if pos < len (file.variations) else ','}''')

    else:
        print (f'''                           0,''')
    print (f'''                           gsk_gpu_clip_get_shader_clip (&pass->clip, &pass->offset, bounds),''')
    print (f'''                           pass->clip_mask,''')
    if file.n_textures > 0:
        print (f'''                           (GskGpuImage *[{file.n_textures}]) {{ {', '.join (map (lambda x: 'image' + str(x), range(1, file.n_textures + 1)))} }},
                           (GskGpuSampler[{file.n_textures}]) {{ {', '.join (map (lambda x: 'sampler' + str(x), range(1, file.n_textures + 1)))} }},''')
    else:
        print (f'''                           NULL,
                           NULL,''')
    print (f'''                           &instance);''')

    for attr in attributes:
        size = 0
        for var in attr.inputs:
            print (var.type.struct_initializer ('  ', 'instance->', var.name if var.name != 'opacity' else 'pass->opacity', attr.name, size))
            size += var.type.size

    print (f'''}}
''')


def print_gl_setup_vao (file, n_attributes, attributes):
    print(f'''static inline void
{file.var_name}_setup_vao (gsize offset)
{{''')

    for attr in attributes:
        for offset in range(0, attr.size, 4):
            size = min (attr.size - offset, 4)
            print(f'''  glEnableVertexAttribArray ({attr.location + offset // 4});
  glVertexAttribDivisor ({attr.location + offset // 4}, 1);
{attr.var_type.gl_attrib_pointer ('  ', attr.location + offset // 4, size, n_attributes * 16, attr.location * 16 + offset * 4)}''')

    print(f'''}}
''');



def print_gl_attrib_locations (file, n_attributes, attributes):
    print(f'''static inline void
{file.var_name}_setup_attrib_locations (GLuint program)
{{''')

    for attr in attributes:
        print(f'''  glBindAttribLocation (program, {attr.location}, "in_{attr.name}");''')

    print(f'''}}
''');


def print_vulkan_info (file, n_attributes, attributes):
    print(f'''#ifdef GDK_RENDERING_VULKAN

static const VkPipelineVertexInputStateCreateInfo {file.var_name}_info = {{
  .sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO,
  .vertexBindingDescriptionCount = 1,
  .pVertexBindingDescriptions = (VkVertexInputBindingDescription[1]) {{
      {{
        .binding = 0,
        .stride = {n_attributes * 16},
        .inputRate = VK_VERTEX_INPUT_RATE_INSTANCE
      }}
  }},
  .vertexAttributeDescriptionCount = {n_attributes},
  .pVertexAttributeDescriptions = (VkVertexInputAttributeDescription[{n_attributes}]) {{''')

    for attr in attributes:
        for offset in range(0, attr.size, 4):
            size = min (attr.size - offset, 4)
            print(f'''      {{
        .location = {attr.location + offset // 4},
        .binding = 0,
        .format = VK_FORMAT_{'R32G32B32A32'[0:size * 3]}_{attr.var_type.vk_type},
        .offset = {attr.location * 16 + offset * 4},
      }},''')

    print(f'''  }},
}};

#endif /* GDK_RENDERING_VULKAN */
''')

def print_glsl_file (file, n_attributes, attributes):
    print (f'''#define GSK_N_TEXTURES {file.n_textures}
{'//#undef' if not file.dual_blend else '#define'} GSK_DUAL_BLEND 1

#include "common.glsl"
''')
    print_glsl_variations (file.variations)
    print_glsl_attributes (attributes)

def print_header_file (file, n_attributes, attributes):
    print (f'''/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"
''')
    print_c_invocation (file, n_attributes, attributes, True)

def print_source_file (file, n_attributes, attributes):
    print (f'''/* This file is auto-generated; changes will not be preserved */

#include "config.h"

#include "gskgpu{file.name}opprivate.h"

#include "gskgpurenderpassprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskrectprivate.h"
#include <graphene.h>

''')
    print_c_struct (file, n_attributes, attributes)
    print_c_print_function (file)
    print_gl_setup_vao (file, n_attributes, attributes)
    print_gl_attrib_locations (file, n_attributes, attributes)
    print_vulkan_info (file, n_attributes, attributes)
    print_c_shader_op_class (file)
    print_c_invocation (file, n_attributes, attributes, False)


parser = argparse.ArgumentParser()
parser.add_argument ('--generate-glsl', action='store_true', help='Generate GLSL includes')
parser.add_argument ('--generate-old-header', action='store_true', help='Generate old version of C header')
parser.add_argument ('--generate-header', action='store_true', help='Generate C header')
parser.add_argument ('--generate-source', action='store_true', help='Generate C source')
parser.add_argument ('FILES', nargs='*', help='Input files')
args = parser.parse_args()

for path in args.FILES:
    file = read_file (path)
    n_attributes, attributes = generate_attributes (file.variables)

    if args.generate_glsl:
        print_glsl_file (file, n_attributes,attributes)
    if args.generate_header:
        print_header_file (file, n_attributes, attributes)
    if args.generate_source:
        print_source_file (file, n_attributes, attributes)




#!/usr/bin/env python3
#
# Generate gdk.gresources.xml
#
# Usage: gen-gdk-gresources-xml SRCDIR_GDK [OUTPUT-FILE]

import os, sys
import filecmp

def replace_if_changed(new, old):
  '''
  Compare contents and only replace if changed to avoid triggering a rebuild.
  '''
  try:
    changed = not filecmp.cmp(new, old, shallow=False)
  except FileNotFoundError:
    changed = True
  if changed:
    os.replace(new, old)
  else:
    os.remove(new)

srcdir = sys.argv[1]

xml = '''<?xml version='1.0' encoding='UTF-8'?>
<gresources>
  <gresource prefix='/org/gtk/libgdk'>
    <file alias='cursor/alias'>cursors/alias_cursor.png</file>
    <file alias='cursor/all-resize'>cursors/all_resize_cursor.png</file>
    <file alias='cursor/all-scroll'>cursors/all_scroll_cursor.png</file>
    <file alias='cursor/cell'>cursors/cell_cursor.png</file>
    <file alias='cursor/col-resize'>cursors/col_resize_cursor.png</file>
    <file alias='cursor/context-menu'>cursors/context_menu_cursor.png</file>
    <file alias='cursor/copy'>cursors/copy_cursor.png</file>
    <file alias='cursor/crosshair'>cursors/crosshair_cursor.png</file>
    <file alias='cursor/default'>cursors/default_cursor.png</file>
    <file alias='cursor/dnd-ask'>cursors/dnd_ask_cursor.png</file>
    <file alias='cursor/e-resize'>cursors/e_resize_cursor.png</file>
    <file alias='cursor/ew-resize'>cursors/ew_resize_cursor.png</file>
    <file alias='cursor/grabbing'>cursors/grabbing_cursor.png</file>
    <file alias='cursor/grab'>cursors/grab_cursor.png</file>
    <file alias='cursor/help'>cursors/help_cursor.png</file>
    <file alias='cursor/move'>cursors/move_cursor.png</file>
    <file alias='cursor/ne-resize'>cursors/ne_resize_cursor.png</file>
    <file alias='cursor/nesw-resize'>cursors/nesw_resize_cursor.png</file>
    <file alias='cursor/no-drop'>cursors/no_drop_cursor.png</file>
    <file alias='cursor/not-allowed'>cursors/not_allowed_cursor.png</file>
    <file alias='cursor/n-resize'>cursors/n_resize_cursor.png</file>
    <file alias='cursor/ns-resize'>cursors/ns_resize_cursor.png</file>
    <file alias='cursor/nw-resize'>cursors/nw_resize_cursor.png</file>
    <file alias='cursor/nwse-resize'>cursors/nwse_resize_cursor.png</file>
    <file alias='cursor/pointer'>cursors/pointer_cursor.png</file>
    <file alias='cursor/progress'>cursors/progress_cursor.png</file>
    <file alias='cursor/row-resize'>cursors/row_resize_cursor.png</file>
    <file alias='cursor/se-resize'>cursors/se_resize_cursor.png</file>
    <file alias='cursor/s-resize'>cursors/s_resize_cursor.png</file>
    <file alias='cursor/sw-resize'>cursors/sw_resize_cursor.png</file>
    <file alias='cursor/text'>cursors/text_cursor.png</file>
    <file alias='cursor/vertical-text'>cursors/vertical_text_cursor.png</file>
    <file alias='cursor/wait'>cursors/wait_cursor.png</file>
    <file alias='cursor/w-resize'>cursors/w_resize_cursor.png</file>
    <file alias='cursor/zoom-in'>cursors/zoom_in_cursor.png</file>
    <file alias='cursor/zoom-out'>cursors/zoom_out_cursor.png</file>
'''

def get_files(subdir,extension):
  return sorted(filter(lambda x: x.endswith((extension)), os.listdir(os.path.join(srcdir,subdir))))

xml += '''
  </gresource>
</gresources>'''

if len(sys.argv) > 2:
  outfile = sys.argv[2]
  tmpfile = outfile + '~'
  with open(tmpfile, 'w') as f:
    f.write(xml)
  replace_if_changed(tmpfile, outfile)
else:
  print(xml)



import sys
import os
import subprocess
import gi

gi.require_version('Gdk', '4.0')

from gi.repository import GLib, Gdk
from pydbus import SessionBus

verbose = True

screen_cast = None
monitors = {}
waiting = False
done = False
monitor_model = None
display = None

def terminate():
    for key in monitors:
        monitor = monitors[key];
        pipeline = monitor['pipeline'];
        pipeline.terminate()
    sys.exit(1)

def stream_added_closure(name):
    def stream_added(node_id):
        if verbose:
            print('pipewire stream added')

        monitor = monitors[name];

        freq = monitor['freq'];
        width = monitor['width'];
        height = monitor['height'];
        # FIXME scale = monitor['scale'];

        # Use gstreamer out-of-process, since the gst gl support gets
        # itself into a twist with its wayland connection when monitors
        # disappear
        pipeline_desc = f'gst-launch-1.0 pipewiresrc path={node_id} ! video/x-raw,max-framerate={freq}/1,width={width},height={height} ! videoconvert ! glimagesink' # >& gstreamer-monitor.log'
        if verbose:
            print(f'launching {pipeline_desc}')
        monitor['pipeline'] = subprocess.Popen([pipeline_desc], shell=True)

    return stream_added

def add_monitor(name, width, height, scale, freq):
    if verbose:
        print(f'add monitor {name}: {width}x{height}, scale {scale}, frequency {freq}')
    session_path = screen_cast.CreateSession({})
    session = bus.get('org.gnome.Mutter.ScreenCast', session_path)
    monitors[name] = {
        'session': session,
        'width': width,
        'height': height,
        'scale': scale,
        'freq': freq
    }
    stream_path = session.RecordVirtual({})
    stream = bus.get('org.gnome.Mutter.ScreenCast', stream_path)
    stream.onPipeWireStreamAdded = stream_added_closure(name)
    session.Start()

def remove_monitor(name):
    if verbose:
        print(f'remove monitor {name}')
    try:
        monitor = monitors[name];
        pipeline = monitor['pipeline']
        pipeline.kill()
        session = monitor['session']
        session.Stop()
    except KeyError:
        print('failed to remove monitor')
    monitors[name] = None

expected_change = None
loop = None

def quit_cb(loop):
    loop.quit()
    print('timed out while waiting')

def wait(millis):
    global loop
    before = GLib.get_monotonic_time()
    loop = GLib.MainLoop()
    GLib.timeout_add(millis, quit_cb, loop)
    loop.run()
    if verbose:
        time = (GLib.get_monotonic_time() - before) / 1000
        print(f'waited for {time} milliseconds')

def monitors_changed(monitors, position, removed, added):
    global expected_change

    assert expected_change != None, 'No change expected'
    assert position == expected_change['position'], 'Unexpected position in monitors-changed'
    assert removed == expected_change['removed'], 'Unexpected removed in monitors-changed'
    assert added == expected_change['added'], 'Unexpected added in monitors-changed'

    if verbose:
        print('got expected monitors-changed signal')

    expected_change = None
    loop.quit()

def launch_observer():
    global monitor_model
    global display

    if display == None:
        display = Gdk.Display.open(os.getenv('WAYLAND_DISPLAY'))

    if verbose:
        print('launch observer')

    monitor_model = display.get_monitors()
    assert monitor_model.get_n_items() == 0, 'Unexpected initial monitors'
    monitor_model.connect('items-changed', monitors_changed)

def expect_monitors_changed(position, removed, added, timeout):
    global expected_change
    expected_change = {
        'position' : position,
        'removed' : removed,
        'added' : added
    }
    wait(timeout)
    assert expected_change == None, 'Expected change did not happen'

def got_connector(monitor, pspec):
    loop.quit()

def expect_monitor(position, width, height, scale, freq):
    assert monitor_model.get_n_items() > position, f'Monitor {position} not present'
    monitor = monitor_model.get_item(position)
    if monitor.get_connector() == None:
        if verbose:
            print('waiting for connector')
        handler = monitor.connect('notify::connector', got_connector)
        wait(500)
        monitor.disconnect(handler)
    assert monitor.get_connector() != None, 'Monitor has no connector'
    assert monitor.is_valid(), 'Monitor is not valid'
    geometry = monitor.get_geometry()
    assert geometry.width == width, 'Unexpected monitor width'
    assert geometry.height == height, 'Unexpected monitor height'
    assert monitor.get_scale_factor() == scale, 'Unexpected scale factor'
    assert monitor.get_refresh_rate() == freq, 'Unexpected monitor frequency'
    if verbose:
        print(f'monitor {position}: {geometry.width}x{geometry.height} frequency {monitor.get_refresh_rate()} scale {monitor.get_scale_factor()} model \'{monitor.get_model()}\' connector \'{monitor.get_connector()}\'')

def run_commands():
    try:
        launch_observer()

        add_monitor('0', width=100, height=100, scale=1, freq=60)
        expect_monitors_changed(0, 0, 1, 10000)
        expect_monitor (position=0, width=100, height=100, scale=1, freq=60000)

        add_monitor('1', width=1024, height=768, scale=1, freq=144)
        expect_monitors_changed(1, 0, 1, 10000)
        expect_monitor (position=1, width=1024, height=768, scale=1, freq=144000)

        remove_monitor('0')
        expect_monitors_changed(0, 1, 0, 11000) # mutter takes 10 seconds to remove it

        remove_monitor('1')
        expect_monitors_changed(0, 1, 0, 11000)
    except AssertionError as e:
        print(f'Error: {e}')
        terminate()

def mutter_appeared(name):
    global screen_cast
    global done
    if verbose:
        print('mutter appeared on the bus')
    screen_cast = bus.get('org.gnome.Mutter.ScreenCast',
                          '/org/gnome/Mutter/ScreenCast')
    run_commands()

    if verbose:
        print ('Done running commands, exiting...')
    done = True

def mutter_vanished():
    global done
    if screen_cast != None:
        if verbose:
            print('mutter left the bus')
        done = True

bus = SessionBus()
bus.watch_name('org.gnome.Mutter.ScreenCast', 0, mutter_appeared, mutter_vanished)

try:
    while not done:
      GLib.MainContext.default().iteration(True)
except KeyboardInterrupt:
    print('Interrupted')



#!/usr/bin/env python3

import argparse
import sys
import re
import os

loaded_files = []

check_defines = [ 'VULKAN', 'GSK_PREAMBLE' ]

def load (path, includes):
    if (path in loaded_files):
        return

    loaded_files.append (path)
    skipping = ''

    with open(path) as f:
        lines = f.readlines()
        for line in lines:
            if skipping:
                match = re.search (r"^#endif /\* (.*) \*/$", line)
                if match and match.group(1) == skipping:
                    skipping = ''
                continue

            match = re.search (r"^#define (.*)$", line)
            if match and match.group(1) in check_defines:
                check_defines.remove (match.group(1))
                print (line, end="")
                continue

            match = re.search (r"^#ifdef (.*)$", line)
            if match and match.group(1) in check_defines:
                skipping = match.group(1)
                continue

            match = re.search (r"^#include \"(.*)\"$", line)
            if match:
                for dir in [os.path.dirname(path)] + includes:
                    file = os.path.join (dir, match.group (1))
                    if not os.path.isfile (file):
                        continue
                    load (file, includes)
                    break
            else:
                print (line, end="")

parser = argparse.ArgumentParser()
parser.add_argument ('-I', '--include', type=str, action='append', help='Add include directory')
parser.add_argument ('FILES', nargs='*', help='Input files')
args = parser.parse_args()

for path in args.FILES:
    load (path, args.include if args.include else [])




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from collections import *


def add_quotes(s):
    return "\"" + s.lower() + "\""


def wordify(s):
    return s.strip().rstrip(".,;:")


def is_keyword(s):
    if s == "GTK":
        return False
    elif s.startswith(("Gtk", "Gdk", "Pango")):
        return True
    elif s.startswith("G") and s[1].isupper():
        return True
    else:
        return False


out_file = sys.argv[1]
in_files = sys.argv[2:]


file_output = """
typedef GtkWidget *(*GDoDemoFunc) (GtkWidget *do_widget);

typedef struct _DemoData DemoData;

struct _DemoData
{
  const char *name;
  const char *title;
  const char **keywords;
  const char *filename;
  GDoDemoFunc func;
  DemoData *children;
};
"""

# Demo = namedtuple('Demo', ['name', 'title', 'keywords', 'file', 'func'])

demos = []

for demo_file in in_files:
    filename =  demo_file[demo_file.rfind('/')+1:]
    demo_name = filename.replace(".c", "")
    with open(demo_file, 'r', encoding='utf-8') as f:
        title = f.readline().replace("/*", "").strip()
        keywords = set()
        line = f.readline().strip();
        while not line.endswith('*/'):
            if line.startswith("* #Keywords:"):
                keywords = keywords.union(set(map(wordify, line.replace ("* #Keywords:", "").strip().split(","))))
            else:
                keywords = keywords.union(set(filter(is_keyword, map(wordify, line.replace ("* ", "").split()))))
            line = f.readline().strip()

    file_output += "GtkWidget *do_" + demo_name + " (GtkWidget *do_widget);\n"
    demos.append((demo_name, title, keywords, filename, "do_" + demo_name, -1))

# Generate a List of "Parent names"
parents = []
parent_ids = []
parent_index = 0
for demo in demos:
    if "/" in demo[1]:
        slash_index = demo[1].index('/')
        parent_name = demo[1][:slash_index]
        do_break = False

        # Check for duplicates
        if not parent_name in parents:
            parents.append(parent_name)
            parent_ids.append(parent_index)
            demos.append(("NULL", parent_name, set(), "NULL", "NULL", parent_index))

        parent_index = parent_index + 1


# Sort demos by title
demos = sorted(demos, key=lambda x: x[1])


# For every child with a parent, generate a list of child demos
i = 0
for parent in parents:
    id = parent_ids[i]
    file_output += "\nDemoData child" + str(id) + "[] = {\n"
    # iterate over all demos and check if the name starts with the given parent name
    for child in demos:
        if child[1].startswith(parent + "/"):
            title = child[1][child[1].rfind('/') + 1:]
            file_output += "  { \"" + child[0] + "\", \"" + title + "\", " + "(const char*[]) {" + ", ".join(list(map(add_quotes, sorted(child[2]))) + ["NULL"]) + " }, \"" + child[3] + "\", " + child[4] + ", NULL },\n"

    file_output += "  { NULL }\n};\n"
    i = i + 1


file_output += "\nDemoData gtk_demos[] = {\n"
for demo in demos:
    # Do not generate one of these for demos with a parent demo
    if "/" not in demo[1]:
        child_array = "NULL"
        name = demo[0]
        title = demo[1]
        keywords = demo[2]
        file = demo[3]
        if name != "NULL":
            name = "\"" + name + "\""
        if title != "NULL":
            title = "\"" + title + "\""
        if file != "NULL":
            file = "\"" + file + "\""

        if demo[5] != -1:
            child_array = "child" + str(demo[5])
        file_output += "  { " + name + ", " + title + ", " + "(const char*[]) {" + ", ".join(list(map(add_quotes, sorted(keywords))) + ["NULL"]) + " }, " + file + ", " + demo[4] + ", " + child_array + " },\n"


file_output += "  { NULL }\n};\n"
ofile = open(out_file, "w")
ofile.write(file_output)
ofile.close()



#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2026 Emmanuele Bassi
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import os
import sys
import xml.etree.ElementTree as ET

from dataclasses import dataclass
from enum import Enum
from pathlib import Path


class TermColor(Enum):
    NONE = 0
    EXT = 1
    TRUE = 2


def setup_output() -> TermColor:
    try:
        if not os.isatty(sys.stdout.fileno()):
            return TermColor.NONE
        if os.environ.get('TERM', '') == 'dumb':
            return TermColor.NONE
        if os.environ.get('COLORTERM', '') == 'truecolor' or \
           os.environ.get('TERM', '') == 'xterm-256color':
            return TermColor.TRUE
        return TermColor.EXT
    except Exception:
        return TermColor.NONE


LOG_COLORIZE_OUTPUT: TermColor = setup_output()


def error(message: str) -> None:
    global LOG_COLORIZE_OUTPUT
    msg = []
    if LOG_COLORIZE_OUTPUT == TermColor.NONE:
        msg.append("ERROR: ")
    elif LOG_COLORIZE_OUTPUT == TermColor.EXT:
        msg.append("\033[1;38;5;160ERROR:\033[0;39;49m ")
    elif LOG_COLORIZE_OUTPUT == TermColor.TRUE:
        msg.append("\033[1;38;2;255;0;0mERROR:\033[0;39;49m ")
    msg.append(message)
    print("".join(msg), file=sys.stderr)
    sys.exit(1)


@dataclass
class Section:
    prefix: str
    files: list[str]


@dataclass
class Manifest:
    file: Path
    sections: list[Section]

    def __str__(self):
        res = [f"Source: {self.file}"]
        for sec in self.sections:
            res.append(f"  Section: {sec.prefix}")
            for f in sec.files:
                res.append(f"    File: {f}")
        return "\n".join(res)


def parse_section(node: ET.Element) -> Section:
    assert node.tag == "gresource"
    prefix = node.attrib.get("prefix", "/")
    files = []
    for child in node:
        if child.tag == "file":
            if child.text is None:
                error("Invalid file component")
            file = child.text
            files.append(file)
    return Section(prefix, files)


def load_manifest(manifest: Path) -> Manifest:
    tree = ET.parse(manifest)
    root = tree.getroot()
    if root.tag != "gresources":
        raise RuntimeError("Invalid manifest file")
    sections = []
    for node in root:
        if node.tag == "gresource":
            section = parse_section(node)
            sections.append(section)
    return Manifest(manifest, sections)


def check_file(manifest: Manifest, file: Path):
    manifest_basedir = manifest.file.parent.name
    # normalize the absolute path to the location of the manifest
    idx = file.parts.index(manifest_basedir)
    check = manifest.file.parent.joinpath(*file.parts[idx + 1:])
    if not check.exists():
        error(f"File {file} not found in {basedir}")
    check_file = "/".join(file.parts[idx + 1:])
    found = False
    for sect in manifest.sections:
        for f in sect.files:
            if f == check_file:
                found = True
                break
        if found:
            break
    if not found:
        error(f"File {file} not found in the manifest")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("manifest", help="The GResource manifest to check")
    parser.add_argument("files", nargs="+", help="The list of files to check")

    args = parser.parse_args()

    try:
        manifest = load_manifest(Path(args.manifest).resolve())
    except Exception as err:
        error(f"Unable to parse {args.manifest}: {err}")

    for file in args.files:
        check_file(manifest, Path(file).resolve())

if __name__ == "__main__":
    main()
